#!/usr/bin/env python
# Copyright (C) 2015 OpenStack, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Manage Jenkins XML config file output.

import hashlib
import pkg_resources
import sys
from xml.dom import minidom
import xml.etree.ElementTree as XML

from jenkins_jobs.errors import JenkinsJobsException

__all__ = ["XmlJobGenerator", "XmlJob"]


def remove_ignorable_whitespace(node):
    """Remove insignificant whitespace from XML nodes

    It should only remove whitespace in between elements and sub elements.
    This should be safe for Jenkins due to how it's XML serialization works
    but may not be valid for other XML documents. So use this method with
    caution outside of this specific library.
    """
    # strip tail whitespace if it's not significant
    if node.tail and node.tail.strip() == "":
        node.tail = None

    for child in node:
        # only strip whitespace from the text node if there are subelement
        # nodes as this means we are removing leading whitespace before such
        # sub elements. Otherwise risk removing whitespace from an element
        # that only contains whitespace
        if node.text and node.text.strip() == "":
            node.text = None
        remove_ignorable_whitespace(child)


class XmlJob(object):
    def __init__(self, xml, name):
        self.xml = xml
        self.name = name

    def md5(self):
        if sys.version_info[:2] >= (3, 6):
            # allows md5 use on fips-enabled systems
            hash_func = hashlib.new("md5", usedforsecurity=False)
            hash_func.update(self.output())
            return hash_func.hexdigest()
        else:
            return hashlib.md5(self.output()).hexdigest()

    def output(self):
        out = minidom.parseString(XML.tostring(self.xml, encoding="UTF-8"))
        return out.toprettyxml(indent="  ", encoding="utf-8")


class XmlGenerator(object):
    """A super-class to capture common XML generation logic.

    Sub-classes should define three attribute: ``entry_point_group``,
    ``kind_attribute`` and ``kind_default``.  The value of ``kind_attribute``
    in the given data dictionary (or ``kind_default`` if it isn't present)
    will be used to filter the entry points in ``entry_point_group``; the
    module so found will be used to generate the XML object.
    """

    def __init__(self, registry):
        self.registry = registry

    def generateXML(self, data_list):
        xml_objs = []
        for data in data_list:
            try:
                xml_objs.append(self._getXMLForData(data.data))
            except JenkinsJobsException as x:
                raise x.with_ctx_list(data.context)
        return xml_objs

    def _getXMLForData(self, data):
        kind = data.get(self.kind_attribute, self.kind_default)

        for ep in pkg_resources.iter_entry_points(
            group=self.entry_point_group, name=kind
        ):
            Mod = ep.load()
            mod = Mod(self.registry)
            xml = mod.root_xml(data)
            if "view-type" not in data:
                self._gen_xml(xml, data)
            obj = XmlJob(xml, data["name"])
            return obj

        names = [
            ep.name
            for ep in pkg_resources.iter_entry_points(group=self.entry_point_group)
        ]
        raise JenkinsJobsException(
            "Unrecognized {}: {} (supported types are: {})".format(
                self.kind_attribute, kind, ", ".join(names)
            )
        )

    def _gen_xml(self, xml, data):
        for module in self.registry.modules:
            if hasattr(module, "gen_xml"):
                module.gen_xml(xml, data)


class XmlJobGenerator(XmlGenerator):
    """Class for generating Jenkins Configuration XML.

    Generates XML from a compatible intermediate representation of Jenkins Jobs.
    """

    entry_point_group = "jenkins_jobs.projects"
    kind_attribute = "project-type"
    kind_default = "freestyle"


class XmlViewGenerator(XmlGenerator):
    """Class for generating Jenkins Configuration XML.

    Generates XML from a compatible intermediate representation of Jenkins Views.
    """

    entry_point_group = "jenkins_jobs.views"
    kind_attribute = "view-type"
    kind_default = "list"



#!/usr/bin/env python
# Copyright (C) 2015 Wayne Warren
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Manage JJB Configuration sources, defaults, and access.

from collections import defaultdict
import io
import logging
import os

from six.moves import configparser, StringIO
from six import PY2

from jenkins_jobs import builder
from jenkins_jobs.errors import JJBConfigException
from jenkins_jobs.errors import JenkinsJobsException

__all__ = ["JJBConfig"]

logger = logging.getLogger(__name__)

DEFAULT_CONF = """
[job_builder]
keep_descriptions=False
ignore_cache=False
recursive=False
exclude=.*
allow_duplicates=False
allow_empty_variables=False
retain_anchors=False
filter_modules=

# other named sections could be used in addition to the implicit [jenkins]
# if you have multiple jenkins servers.
[jenkins]
url=http://localhost:8080/
query_plugins_info=False
"""

CONFIG_REQUIRED_MESSAGE = (
    "A valid configuration file is required. " "No configuration file passed."
)
DEPRECATED_PLUGIN_CONFIG_SECTION_MESSAGE = (
    "Defining plugin configuration using a [{plugin}] section in your config"
    " file is deprecated. The recommended way to define plugins now is by"
    ' using a [plugin "{plugin}"] section'
)
_NOTSET = object()


class JJBConfig(object):
    def __init__(
        self, config_filename=None, config_file_required=False, config_section="jenkins"
    ):

        """
        The JJBConfig class is intended to encapsulate and resolve priority
        between all sources of configuration for the JJB library. This allows
        the various sources of configuration to provide a consistent accessor
        interface regardless of where they are used.

        It also allows users of JJB-as-an-API to create minimally valid
        configuration and easily make minor modifications to default values
        without strictly adhering to the confusing setup (see the _setup
        method, the behavior of which largely lived in the cmd.execute method
        previously) necessary for the jenkins-jobs command line tool.

        :arg str config_filename: Name of configuration file on which to base
            this config object.
        :arg bool config_file_required: Allows users of the JJBConfig class to
            decide whether or not it's really necessary for a config file to be
            passed in when creating an instance. This has two effects on the
            behavior of JJBConfig initialization:
            * It determines whether or not we try "local" and "global" config
              files.
            * It determines whether or not failure to read some config file
              will raise an exception or simply print a warning message
              indicating that no config file was found.
        """

        config_parser = self._init_defaults()

        global_conf = "/etc/jenkins_jobs/jenkins_jobs.ini"
        user_conf = os.path.join(
            os.path.expanduser("~"), ".config", "jenkins_jobs", "jenkins_jobs.ini"
        )
        local_conf = os.path.join(os.path.dirname(__file__), "jenkins_jobs.ini")
        conf = None
        if config_filename is not None:
            conf = config_filename
        else:
            if os.path.isfile(local_conf):
                conf = local_conf
            elif os.path.isfile(user_conf):
                conf = user_conf
            else:
                conf = global_conf

        if config_file_required and conf is None:
            raise JJBConfigException(CONFIG_REQUIRED_MESSAGE)

        config_fp = None
        if conf is not None:
            try:
                config_fp = self._read_config_file(conf)
            except JJBConfigException:
                if config_file_required:
                    raise JJBConfigException(CONFIG_REQUIRED_MESSAGE)
                else:
                    logger.warning(
                        "Config file, {0}, not found. Using "
                        "default config values.".format(conf)
                    )

        if config_fp is not None:
            if PY2:
                config_parser.readfp(config_fp)
            else:
                config_parser.read_file(config_fp)

        self.config_parser = config_parser

        self._section = config_section
        self.print_job_urls = False

        self.jenkins = defaultdict(None)
        self.builder = defaultdict(None)
        self.yamlparser = defaultdict(None)

        self._setup()
        self._handle_deprecated_hipchat_config()

        if config_fp is not None:
            config_fp.close()

    def _init_defaults(self):
        """Initialize default configuration values using DEFAULT_CONF."""
        config = configparser.ConfigParser()
        # Load default config always
        if PY2:
            config.readfp(StringIO(DEFAULT_CONF))
        else:
            config.read_file(StringIO(DEFAULT_CONF))
        return config

    def _read_config_file(self, config_filename):
        """Return a ConfigParser object from a file input."""
        if os.path.isfile(config_filename):
            self.__config_file = config_filename  # remember file we read from
            logger.debug("Reading config from {0}".format(config_filename))
            config_fp = io.open(config_filename, "r", encoding="utf-8")
        else:
            raise JJBConfigException(
                "A valid configuration file is required. "
                "\n{0} is not valid.".format(config_filename)
            )

        return config_fp

    def _handle_deprecated_hipchat_config(self):
        config = self.config_parser

        if config.has_section("hipchat"):
            if config.has_section('plugin "hipchat"'):
                logger.warning(
                    'Both [hipchat] and [plugin "hipchat"] sections '
                    "defined, legacy [hipchat] section will be ignored."
                )
            else:
                logger.warning(
                    "[hipchat] section is deprecated and should be moved to a "
                    '[plugins "hipchat"] section instead as the [hipchat] '
                    "section will be ignored in the future."
                )
                config.add_section('plugin "hipchat"')
                for option in config.options("hipchat"):
                    config.set(
                        'plugin "hipchat"', option, config.get("hipchat", option)
                    )

                config.remove_section("hipchat")

        # remove need to reference jenkins section when using hipchat plugin
        # moving to backports configparser would allow use of extended
        # interpolation to remove the need for plugins to need information
        # directly from the jenkins section within code and allow variables
        # in the config file to refer instead.
        if config.has_section('plugin "hipchat"') and not config.has_option(
            'plugin "hipchat"', "url"
        ):
            config.set('plugin "hipchat"', "url", config.get("jenkins", "url"))

    def _setup(self):
        config = self.config_parser

        logger.debug("Config: {0}".format(config))

        # check the ignore_cache setting
        ignore_cache = False
        if config.has_option(self._section, "ignore_cache"):
            logger.warning(
                "ignore_cache option should be moved to the "
                "[job_builder] section in the config file, the "
                "one specified in the [jenkins] section will be "
                "ignored in the future"
            )
            ignore_cache = config.getboolean(self._section, "ignore_cache")
        elif config.has_option("job_builder", "ignore_cache"):
            ignore_cache = config.getboolean("job_builder", "ignore_cache")
        self.builder["ignore_cache"] = ignore_cache

        # check the flush_cache setting
        flush_cache = False
        if config.has_option("job_builder", "flush_cache"):
            flush_cache = config.getboolean("job_builder", "flush_cache")
        self.builder["flush_cache"] = flush_cache

        # check the print_job_urls setting
        if config.has_option("job_builder", "print_job_urls"):
            self.print_job_urls = config.getboolean("job_builder", "print_job_urls")

        # Jenkins supports access as an anonymous user, which can be used to
        # ensure read-only behaviour when querying the version of plugins
        # installed for test mode to generate XML output matching what will be
        # uploaded. To enable must pass 'None' as the value for user and
        # password to python-jenkins
        #
        # catching 'TypeError' is a workaround for python 2.6 interpolation
        # error
        # https://bugs.launchpad.net/openstack-ci/+bug/1259631

        try:
            user = config.get(self._section, "user")
        except (TypeError, configparser.NoOptionError):
            user = None
        self.jenkins["user"] = user

        try:
            password = config.get(self._section, "password")
        except (TypeError, configparser.NoOptionError):
            password = None
        self.jenkins["password"] = password

        # None -- no timeout, blocking mode; same as setblocking(True)
        # 0.0 -- non-blocking mode; same as setblocking(False) <--- default
        # > 0 -- timeout mode; operations time out after timeout seconds
        # < 0 -- illegal; raises an exception
        # to retain the default must use
        # "timeout=jenkins_jobs.builder._DEFAULT_TIMEOUT" or not set timeout at
        # all.
        try:
            timeout = config.getfloat(self._section, "timeout")
        except (ValueError):
            raise JenkinsJobsException("Jenkins timeout config is invalid")
        except (TypeError, configparser.NoOptionError):
            timeout = builder._DEFAULT_TIMEOUT
        self.jenkins["timeout"] = timeout

        plugins_info = None
        if config.has_option(
            self._section, "query_plugins_info"
        ) and not config.getboolean(self._section, "query_plugins_info"):
            logger.debug("Skipping plugin info retrieval")
            plugins_info = []
        self.builder["plugins_info"] = plugins_info

        self.recursive = config.getboolean("job_builder", "recursive")
        self.excludes = config.get("job_builder", "exclude").split(os.pathsep)

        # The way we want to do things moving forward:
        self.jenkins["url"] = config.get(self._section, "url")
        self.builder["print_job_urls"] = self.print_job_urls

        # keep descriptions ? (used by yamlparser)
        keep_desc = False
        if (
            config
            and config.has_section("job_builder")
            and config.has_option("job_builder", "keep_descriptions")
        ):
            keep_desc = config.getboolean("job_builder", "keep_descriptions")
        self.yamlparser["keep_descriptions"] = keep_desc

        # figure out the include path (used by yamlparser)
        path = ["."]
        if (
            config
            and config.has_section("job_builder")
            and config.has_option("job_builder", "include_path")
        ):
            path = config.get("job_builder", "include_path").split(":")
        self.yamlparser["include_path"] = path

        # Extra modules to load for jinja2 filters
        filter_modules = []
        if (
            config
            and config.has_section("job_builder")
            and config.has_option("job_builder", "filter_modules")
            and config.get("job_builder", "filter_modules")
        ):
            filter_modules = config.get("job_builder", "filter_modules").split(" ")
        self.yamlparser["filter_modules"] = filter_modules

        # allow duplicates?
        allow_duplicates = False
        if config and config.has_option("job_builder", "allow_duplicates"):
            allow_duplicates = config.getboolean("job_builder", "allow_duplicates")
        self.yamlparser["allow_duplicates"] = allow_duplicates

        # allow empty variables?
        self.yamlparser["allow_empty_variables"] = (
            config
            and config.has_section("job_builder")
            and config.has_option("job_builder", "allow_empty_variables")
            and config.getboolean("job_builder", "allow_empty_variables")
        )

        # retain anchors across files?
        retain_anchors = False
        if config and config.has_option("job_builder", "retain_anchors"):
            retain_anchors = config.getboolean("job_builder", "retain_anchors")
        self.yamlparser["retain_anchors"] = retain_anchors

        update = None
        if (
            config
            and config.has_section("job_builder")
            and config.has_option("job_builder", "update")
        ):
            update = config.get("job_builder", "update")
        self.builder["update"] = update

    def validate(self):
        # Inform the user as to what is likely to happen, as they may specify
        # a real jenkins instance in test mode to get the plugin info to check
        # the XML generated.
        if self.jenkins["user"] is None and self.jenkins["password"] is None:
            logger.info("Will use anonymous access to Jenkins if needed.")
        elif (
            self.jenkins["user"] is not None and self.jenkins["password"] is None
        ) or (self.jenkins["user"] is None and self.jenkins["password"] is not None):
            raise JenkinsJobsException(
                "Cannot authenticate to Jenkins with only one of User and "
                "Password provided, please check your configuration."
            )

        if self.builder["plugins_info"] is not None and not isinstance(
            self.builder["plugins_info"], list
        ):
            raise JenkinsJobsException("plugins_info must contain a list!")

    def get_module_config(self, section, key, default=None):
        """Returns the value of a config in a config module.

        Given a section name and a key value, return the value assigned to
        the key in the JJB .ini file if it exists, otherwise emit a warning
        indicating that the value is not set. Default value returned if no
        value is set in the file will be a blank string.
        """
        result = default
        try:
            result = self.config_parser.get(section, key)
        except (
            configparser.NoSectionError,
            configparser.NoOptionError,
            JenkinsJobsException,
        ) as e:
            # use of default ignores missing sections/options
            if result is None:
                logger.warning(
                    "You didn't set a %s neither in the yaml job definition "
                    "nor in the %s section, blank default value will be "
                    "applied:\n%s",
                    key,
                    section,
                    e,
                )
        return result

    def get_plugin_config(self, plugin, key, default=None):
        return self.get_module_config('plugin "{}"'.format(plugin), key, default)



# Joint copyright:
#  - Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# The goal of these tests is to check that given a particular set of flags to
# Jenkins Job Builder's command line tools it will result in a particular set
# of actions by the JJB library, usually through interaction with the
# python-jenkins library.

from unittest import mock

import pytest


def test_update_jobs(mocker, fixtures_dir, default_config_file, execute_jenkins_jobs):
    """
    Test update_job is called
    """
    mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.job_exists")
    mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.get_all_jobs")
    mocker.patch(
        "jenkins_jobs.builder.jenkins.Jenkins.get_job_info",
        return_value={"disabled": False},
    )
    reconfig_job = mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.reconfig_job")

    path = fixtures_dir / "cmd-002.yaml"
    args = ["--conf", default_config_file, "update", str(path)]

    execute_jenkins_jobs(args)

    reconfig_job.assert_has_calls(
        [
            mock.call(job_name, mock.ANY)
            for job_name in ["bar001", "bar002", "baz001", "bam001"]
        ],
        any_order=True,
    )


def test_update_jobs_enabled_only(
    mocker, fixtures_dir, default_config_file, execute_jenkins_jobs
):
    """
    Test update_job is called with --enabled-only
    """
    mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.job_exists")
    mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.get_all_jobs")
    mocker.patch(
        "jenkins_jobs.builder.jenkins.Jenkins.get_job_info",
        return_value={"disabled": False},
    )
    reconfig_job = mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.reconfig_job")

    path = fixtures_dir / "cmd-002.yaml"
    args = ["--conf", default_config_file, "update", "--enabled-only", str(path)]

    execute_jenkins_jobs(args)

    reconfig_job.assert_has_calls(
        [mock.call(job_name, mock.ANY) for job_name in ["bar001", "bar002", "baz001"]],
        any_order=True,
    )
    # make sure that there are only those 3 calls checked before
    assert len(reconfig_job.call_args_list) == 3


def test_update_jobs_decode_job_output(
    mocker, fixtures_dir, default_config_file, execute_jenkins_jobs
):
    """
    Test that job xml output has been decoded before attempting to update
    """
    mocker.patch("jenkins_jobs.builder.JenkinsManager.is_job", return_value=True)
    mocker.patch("jenkins_jobs.builder.JenkinsManager.get_jobs")
    mocker.patch("jenkins_jobs.builder.JenkinsManager.get_job_md5")
    mocker.patch(
        "jenkins_jobs.builder.jenkins.Jenkins.get_job_info",
        return_value={"disabled": False},
    )
    update_job_mock = mocker.patch("jenkins_jobs.builder.JenkinsManager.update_job")

    # don't care about the value returned here
    update_job_mock.return_value = ([], 0)

    path = fixtures_dir / "cmd-002.yaml"
    args = ["--conf", default_config_file, "update", str(path)]

    execute_jenkins_jobs(args)
    assert isinstance(update_job_mock.call_args[0][1], str)


def test_update_jobs_and_delete_old(
    mocker, fixtures_dir, default_config_file, execute_jenkins_jobs
):
    """Test update behaviour with --delete-old option.

    * mock out a call to jenkins.Jenkins.get_jobs() to return a known list
      of job names.
    * mock out a call to jenkins.Jenkins.reconfig_job() and
      jenkins.Jenkins.delete_job() to detect calls being made to determine
      that JJB does correctly delete the jobs it should delete when passed
      a specific set of inputs.
    * mock out a call to jenkins.Jenkins.job_exists() to always return
      True.
    """
    mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.job_exists")
    mocker.patch(
        "jenkins_jobs.builder.jenkins.Jenkins.get_job_info",
        return_value={"disabled": False},
    )
    jenkins_get_all_jobs = mocker.patch(
        "jenkins_jobs.builder.jenkins.Jenkins.get_all_jobs"
    )
    jenkins_reconfig_job = mocker.patch(
        "jenkins_jobs.builder.jenkins.Jenkins.reconfig_job"
    )
    jenkins_delete_job = mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.delete_job")

    mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.get_views")

    yaml_jobs = ["bar001", "bar002", "baz001", "bam001"]
    extra_jobs = ["old_job001", "old_job002", "unmanaged"]

    path = fixtures_dir / "cmd-002.yaml"
    args = ["--conf", default_config_file, "update", "--delete-old", str(path)]

    jenkins_get_all_jobs.return_value = [
        {"fullname": name} for name in yaml_jobs + extra_jobs
    ]

    mocker.patch(
        "jenkins_jobs.builder.JenkinsManager.is_managed_job",
        side_effect=(lambda name: name != "unmanaged"),
    )
    execute_jenkins_jobs(args)

    jenkins_reconfig_job.assert_has_calls(
        [mock.call(job_name, mock.ANY) for job_name in yaml_jobs], any_order=True
    )
    calls = [mock.call(name) for name in extra_jobs if name != "unmanaged"]
    jenkins_delete_job.assert_has_calls(calls)
    # to ensure only the calls we expected were made, have to check
    # there were no others, as no API call for assert_has_only_calls
    assert jenkins_delete_job.call_count == len(calls)


def test_update_jobs_and_delete_old_views_only(
    mocker, fixtures_dir, default_config_file, execute_jenkins_jobs
):
    """Test update behaviour with --delete-old option
    with --views-only option specified.
    No jobs should be deleted.
    """
    mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.job_exists")
    jenkins_get_all_jobs = mocker.patch(
        "jenkins_jobs.builder.jenkins.Jenkins.get_all_jobs"
    )
    jenkins_reconfig_job = mocker.patch(
        "jenkins_jobs.builder.jenkins.Jenkins.reconfig_job"
    )
    jenkins_delete_job = mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.delete_job")

    mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.view_exists")
    jenkins_get_all_views = mocker.patch(
        "jenkins_jobs.builder.jenkins.Jenkins.get_views"
    )
    jenkins_reconfig_view = mocker.patch(
        "jenkins_jobs.builder.jenkins.Jenkins.reconfig_view"
    )
    jenkins_delete_view = mocker.patch(
        "jenkins_jobs.builder.jenkins.Jenkins.delete_view"
    )

    yaml_jobs = ["job-1", "job-2", "job-3"]
    extra_managed_jobs = ["old-job-1", "old-job-2"]
    unmanaged_jobs = ["unmanaged-job"]

    yaml_views = ["view-1", "view-2", "view-3"]
    extra_managed_views = ["old-view-1", "old-view-2"]
    unmanaged_views = ["unmanaged-view"]

    jenkins_get_all_jobs.return_value = [
        {"fullname": name} for name in yaml_jobs + extra_managed_jobs + unmanaged_jobs
    ]
    jenkins_get_all_views.return_value = [
        {"name": name} for name in yaml_views + extra_managed_views + unmanaged_views
    ]

    mocker.patch(
        "jenkins_jobs.builder.JenkinsManager.is_managed_job",
        side_effect=(lambda name: name not in unmanaged_jobs),
    )
    mocker.patch(
        "jenkins_jobs.builder.JenkinsManager.is_managed_view",
        side_effect=(lambda name: name not in unmanaged_views),
    )

    path = fixtures_dir / "update-both.yaml"
    args = [
        "--conf",
        default_config_file,
        "update",
        "--delete-old",
        "--views-only",
        str(path),
    ]

    execute_jenkins_jobs(args)

    assert jenkins_reconfig_job.call_count == 0
    assert jenkins_delete_job.call_count == 0

    assert jenkins_reconfig_view.call_count == 3
    assert jenkins_delete_view.call_count == 2


def test_update_views(mocker, fixtures_dir, default_config_file, execute_jenkins_jobs):
    """
    Test update_job is called for project with views
    """
    mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.view_exists")
    mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.get_views")
    reconfig_view = mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.reconfig_view")

    path = fixtures_dir / "update-views.yaml"
    args = ["--conf", default_config_file, "update", str(path)]

    execute_jenkins_jobs(args)

    reconfig_view.assert_has_calls(
        [
            mock.call(view_name, mock.ANY)
            for view_name in ["view-1", "view-2", "view-3"]
        ],
        any_order=True,
    )
    assert reconfig_view.call_count == 3


def test_update_jobs_and_views(
    mocker, fixtures_dir, default_config_file, execute_jenkins_jobs
):
    """
    Test update_job is called for project with both jobs and views
    """
    mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.job_exists")
    mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.get_all_jobs")
    mocker.patch(
        "jenkins_jobs.builder.jenkins.Jenkins.get_job_info",
        return_value={"disabled": False},
    )
    reconfig_job = mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.reconfig_job")

    mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.view_exists")
    mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.get_views")
    reconfig_view = mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.reconfig_view")

    path = fixtures_dir / "update-both.yaml"
    args = ["--conf", default_config_file, "update", str(path)]

    execute_jenkins_jobs(args)

    reconfig_job.assert_has_calls(
        [mock.call(job_name, mock.ANY) for job_name in ["job-1", "job-2", "job-3"]],
        any_order=True,
    )
    assert reconfig_job.call_count == 3

    reconfig_view.assert_has_calls(
        [
            mock.call(view_name, mock.ANY)
            for view_name in ["view-1", "view-2", "view-3"]
        ],
        any_order=True,
    )
    assert reconfig_view.call_count == 3


def test_update_jobs_only(
    mocker, fixtures_dir, default_config_file, execute_jenkins_jobs
):
    """
    Test update_job is called for project with both jobs and views
    but only jobs update is requested
    """
    mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.job_exists")
    mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.get_all_jobs")
    mocker.patch(
        "jenkins_jobs.builder.jenkins.Jenkins.get_job_info",
        return_value={"disabled": False},
    )
    reconfig_job = mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.reconfig_job")

    mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.view_exists")
    mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.get_views")
    reconfig_view = mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.reconfig_view")

    path = fixtures_dir / "update-both.yaml"
    args = ["--conf", default_config_file, "update", "--jobs-only", str(path)]

    execute_jenkins_jobs(args)

    assert reconfig_job.call_count == 3
    assert reconfig_view.call_count == 0


def test_update_views_only(
    mocker, fixtures_dir, default_config_file, execute_jenkins_jobs
):
    """
    Test update_job is called for project with both jobs and views
    but only views update is requested
    """
    mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.job_exists")
    mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.get_all_jobs")
    reconfig_job = mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.reconfig_job")

    mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.view_exists")
    mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.get_views")
    reconfig_view = mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.reconfig_view")

    path = fixtures_dir / "update-both.yaml"
    args = ["--conf", default_config_file, "update", "--views-only", str(path)]

    execute_jenkins_jobs(args)

    assert reconfig_job.call_count == 0
    assert reconfig_view.call_count == 3


def test_update_views_and_delete_old(
    mocker, fixtures_dir, default_config_file, execute_jenkins_jobs
):
    """Test update behaviour with --delete-old option for views."""
    mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.get_all_jobs")
    mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.view_exists")
    jenkins_get_all_views = mocker.patch(
        "jenkins_jobs.builder.jenkins.Jenkins.get_views"
    )
    jenkins_reconfig_view = mocker.patch(
        "jenkins_jobs.builder.jenkins.Jenkins.reconfig_view"
    )
    jenkins_delete_view = mocker.patch(
        "jenkins_jobs.builder.jenkins.Jenkins.delete_view"
    )

    yaml_views = ["view-1", "view-2", "view-3"]
    extra_managed_views = ["old-view-1", "old-view-2"]
    unmanaged_views = ["unmanaged"]

    path = fixtures_dir / "update-views.yaml"
    args = ["--conf", default_config_file, "update", "--delete-old", str(path)]

    jenkins_get_all_views.return_value = [
        {"name": name} for name in yaml_views + extra_managed_views + unmanaged_views
    ]

    mocker.patch(
        "jenkins_jobs.builder.JenkinsManager.is_managed_view",
        side_effect=(lambda name: name not in unmanaged_views),
    )
    execute_jenkins_jobs(args)

    jenkins_reconfig_view.assert_has_calls(
        [mock.call(view_name, mock.ANY) for view_name in yaml_views], any_order=True
    )
    calls = [mock.call(name) for name in extra_managed_views]
    jenkins_delete_view.assert_has_calls(calls)
    assert jenkins_delete_view.call_count == len(calls)


def test_update_views_and_delete_old_jobs_only(
    mocker, fixtures_dir, default_config_file, execute_jenkins_jobs
):
    """Test update behaviour with --delete-old option
    with --jobs-only option specified.
    No views should be deleted or updated.
    """
    mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.job_exists")
    mocker.patch(
        "jenkins_jobs.builder.jenkins.Jenkins.get_job_info",
        return_value={"disabled": False},
    )
    jenkins_get_all_jobs = mocker.patch(
        "jenkins_jobs.builder.jenkins.Jenkins.get_all_jobs"
    )
    jenkins_reconfig_job = mocker.patch(
        "jenkins_jobs.builder.jenkins.Jenkins.reconfig_job"
    )
    jenkins_delete_job = mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.delete_job")

    mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.view_exists")
    jenkins_get_all_views = mocker.patch(
        "jenkins_jobs.builder.jenkins.Jenkins.get_views"
    )
    jenkins_reconfig_view = mocker.patch(
        "jenkins_jobs.builder.jenkins.Jenkins.reconfig_view"
    )
    jenkins_delete_view = mocker.patch(
        "jenkins_jobs.builder.jenkins.Jenkins.delete_view"
    )

    yaml_jobs = ["job-1", "job-2", "job-3"]
    extra_managed_jobs = ["old-job-1", "old-job-2"]
    unmanaged_jobs = ["unmanaged-job"]

    yaml_views = ["view-1", "view-2", "view-3"]
    extra_managed_views = ["old-view-1", "old-view-2"]
    unmanaged_views = ["unmanaged-view"]

    path = fixtures_dir / "update-both.yaml"
    args = [
        "--conf",
        default_config_file,
        "update",
        "--delete-old",
        "--jobs-only",
        str(path),
    ]

    jenkins_get_all_jobs.return_value = [
        {"fullname": name} for name in yaml_jobs + extra_managed_jobs + unmanaged_jobs
    ]
    jenkins_get_all_views.return_value = [
        {"name": name} for name in yaml_views + extra_managed_views + unmanaged_views
    ]

    mocker.patch(
        "jenkins_jobs.builder.JenkinsManager.is_managed_job",
        side_effect=(lambda name: name not in unmanaged_jobs),
    )
    mocker.patch(
        "jenkins_jobs.builder.JenkinsManager.is_managed_view",
        side_effect=(lambda name: name not in unmanaged_views),
    )

    execute_jenkins_jobs(args)

    assert jenkins_reconfig_job.call_count == 3
    assert jenkins_delete_job.call_count == 2

    assert jenkins_reconfig_view.call_count == 0
    assert jenkins_delete_view.call_count == 0


def test_update_jobs_fix_disabled(
    mocker, fixtures_dir, default_config_file, execute_jenkins_jobs
):
    """
    Test update_job fix_disabled
    """

    # Case 1: Explicitly specifying disabled as True or False will skip the fix_disabled method.
    mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.job_exists")
    mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.get_all_jobs")
    jenkins_get_job_info = mocker.patch(
        "jenkins_jobs.builder.jenkins.Jenkins.get_job_info",
        return_value={"disabled": False},
    )
    reconfig_job = mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.reconfig_job")

    path = fixtures_dir / "cmd-002.yaml"
    args = ["--conf", default_config_file, "update", str(path)]

    execute_jenkins_jobs(args)

    jenkins_get_job_info.assert_has_calls(
        [mock.call(job_name) for job_name in ["bar001"]],
        any_order=True,
    )
    assert jenkins_get_job_info.call_count == 1

    # Case 2: allow disabled to remain the last setting
    reconfig_job.reset_mock()

    def check_disabled_value(xml_str, expected_value):
        import xml.etree.ElementTree as XML

        xml = XML.fromstring(
            xml_str,
        )
        disabled = xml.find("disabled")
        if expected_value is None:
            assert disabled is None
        else:
            assert disabled is not None
            assert disabled.text == str(expected_value).lower()

    path = fixtures_dir / "cmd-001.yaml"
    args = ["--conf", default_config_file, "update", str(path)]

    jenkins_get_job_info.return_value = {}
    execute_jenkins_jobs(args)
    check_disabled_value(reconfig_job.call_args[0][1], False)

    jenkins_get_job_info.return_value = {"disabled": False}
    execute_jenkins_jobs(args)
    check_disabled_value(reconfig_job.call_args[0][1], False)

    jenkins_get_job_info.return_value = {"disabled": True}
    execute_jenkins_jobs(args)
    check_disabled_value(reconfig_job.call_args[0][1], True)

    # Case 3: if job does not exist, then the value of disabled depends on the job definition
    mocker.patch("jenkins_jobs.builder.JenkinsManager.is_job", return_value=False)
    create_job = mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.create_job")

    path = fixtures_dir / "cmd-002.yaml"
    args = ["--conf", default_config_file, "update", str(path)]

    create_job.reset_mock()
    execute_jenkins_jobs(args)

    expected_values = {
        "bar001": None,
        "bar002": False,
        "baz001": False,
        "bam001": True,
    }
    assert create_job.call_count == 4
    for call in create_job.call_args_list:
        check_disabled_value(call[0][1], expected_values[call[0][0]])


@pytest.mark.skip(reason="TODO: Develop actual update timeout test approach.")
def test_update_timeout_not_set():
    """Validate update timeout behavior when timeout not explicitly configured."""
    pass


@pytest.mark.skip(reason="TODO: Develop actual update timeout test approach.")
def test_update_timeout_set():
    """Validate update timeout behavior when timeout is explicitly configured."""
    pass



# Copyright 2012 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""
Reporters are like publishers but only applicable to Maven projects.

**Component**: reporters
  :Macro: reporter
  :Entry Point: jenkins_jobs.reporters

Example::

  job:
    name: test_job
    project-type: maven

    reporters:
      - email:
          recipients: breakage@example.com
"""

import xml.etree.ElementTree as XML

from jenkins_jobs.errors import JenkinsJobsException
import jenkins_jobs.modules.base
import jenkins_jobs.modules.helpers as helpers


def email(registry, xml_parent, data):
    """yaml: email
    Email notifications on build failure.

    :arg str recipients: Recipient email addresses
    :arg bool notify-every-unstable-build: Send an email for every
      unstable build (default true)
    :arg bool send-to-individuals: Send an email to the individual
      who broke the build (default false)
    :arg bool notify-for-each-module: Send an email for each module
      (e.g. failed, unstable). (default true)

    Example::

      reporters:
        - email:
            recipients: breakage@example.com
    """

    mailer = XML.SubElement(xml_parent, "hudson.maven.reporters.MavenMailer")
    XML.SubElement(mailer, "recipients").text = data["recipients"]

    # Note the logic reversal (included here to match the GUI
    if data.get("notify-every-unstable-build", True):
        XML.SubElement(mailer, "dontNotifyEveryUnstableBuild").text = "false"
    else:
        XML.SubElement(mailer, "dontNotifyEveryUnstableBuild").text = "true"
    mapping = [
        ("send-to-individuals", "sendToIndividuals", False),
        ("notify-for-each-module", "perModuleEmail", True),
    ]
    helpers.convert_mapping_to_xml(mailer, data, mapping, fail_required=False)


def findbugs(registry, xml_parent, data):
    """yaml: findbugs
    FindBugs reporting for builds

    Requires the Jenkins FindBugs Plugin
    (:jenkins-github:`findbugs-plugin`).

    :arg bool rank-priority: Use rank as priority (default false)
    :arg str include-files: Comma separated list of files to include.
        (Optional)
    :arg str exclude-files: Comma separated list of files to exclude.
        (Optional)
    :arg bool can-run-on-failed: Weather or not to run plug-in on failed builds
        (default false)
    :arg int healthy: Sunny threshold (optional)
    :arg int unhealthy: Stormy threshold (optional)
    :arg str health-threshold: Threshold priority for health status
        ('low', 'normal' or 'high', defaulted to 'low')
    :arg bool dont-compute-new: If set to false, computes new warnings based on
        the reference build (default true)
    :arg bool use-delta-values: Use delta for new warnings. (default false)
    :arg bool use-previous-build-as-reference:  If set then the number of new
        warnings will always be calculated based on the previous build.
        Otherwise the reference build. (default false)
    :arg bool use-stable-build-as-reference: The number of new warnings will be
        calculated based on the last stable build, allowing reverts of unstable
        builds where the number of warnings was decreased. (default false)
    :arg dict thresholds:
        :thresholds:
            * **unstable** (`dict`)
                :unstable: * **total-all** (`int`)
                           * **total-high** (`int`)
                           * **total-normal** (`int`)
                           * **total-low** (`int`)
                           * **new-all** (`int`)
                           * **new-high** (`int`)
                           * **new-normal** (`int`)
                           * **new-low** (`int`)

            * **failed** (`dict`)
                :failed: * **total-all** (`int`)
                         * **total-high** (`int`)
                         * **total-normal** (`int`)
                         * **total-low** (`int`)
                         * **new-all** (`int`)
                         * **new-high** (`int`)
                         * **new-normal** (`int`)
                         * **new-low** (`int`)

    Minimal Example:

    .. literalinclude::  /../../tests/reporters/fixtures/findbugs-minimal.yaml

    Full Example:

    .. literalinclude::  /../../tests/reporters/fixtures/findbugs01.yaml
    """
    findbugs = XML.SubElement(xml_parent, "hudson.plugins.findbugs.FindBugsReporter")
    findbugs.set("plugin", "findbugs")

    helpers.findbugs_settings(findbugs, data)
    helpers.build_trends_publisher("[FINDBUGS] ", findbugs, data)


class Reporters(jenkins_jobs.modules.base.Base):
    sequence = 55

    component_type = "reporter"
    component_list_type = "reporters"

    def gen_xml(self, xml_parent, data):
        if "reporters" not in data:
            return

        if xml_parent.tag != "maven2-moduleset":
            raise JenkinsJobsException(
                "Reporters may only be used for Maven " "modules."
            )

        xml_reporters = XML.SubElement(xml_parent, "reporters")

        component_list = data.get("reporters", [])
        self.dispatch_component_list("reporter", component_list, xml_reporters)



# Copyright 2019 Openstack Foundation

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The view nested module handles creating Jenkins Nested views.

To create a nested view specify ``nested`` in the ``view-type`` attribute
to the :ref:`view_nested` definition.

:View Parameters:
    * **name** (`str`): The name of the view.
    * **view-type** (`str`): The type of view.
    * **description** (`str`): A description of the view. (default '')
    * **filter-executors** (`bool`): Show only executors that can
      execute the included views. (default false)
    * **filter-queue** (`bool`): Show only included jobs in builder
      queue. (default false)
    * **views** (`list`): The views to nest.
    * **default-view** (`str`): Name of the view to use as the default from the
      nested ones. (the first one by default)
    * **columns** (`list`): List of columns to be shown in view. (default empty
      list)

Example:

    .. literalinclude::
        /../../tests/views/fixtures/view_nested.yaml

"""

import xml.etree.ElementTree as XML
import jenkins_jobs.modules.base
import jenkins_jobs.modules.helpers as helpers
from jenkins_jobs.root_base import JobViewData
from jenkins_jobs.xml_config import XmlViewGenerator

COLUMN_DICT = {
    "status": "hudson.views.StatusColumn",
    "weather": "hudson.views.WeatherColumn",
}


class Nested(jenkins_jobs.modules.base.Base):
    def root_xml(self, data):
        root = XML.Element("hudson.plugins.nested__view.NestedView")

        mapping = [
            ("name", "name", None),
            ("description", "description", ""),
            ("filter-executors", "filterExecutors", False),
            ("filter-queue", "filterQueue", False),
        ]
        helpers.convert_mapping_to_xml(root, data, mapping, fail_required=True)

        XML.SubElement(root, "properties", {"class": "hudson.model.View$PropertyList"})

        v_xml = XML.SubElement(root, "views")
        views = data.get("views", [])
        view_data_list = [JobViewData(v) for v in views]

        xml_view_generator = XmlViewGenerator(self.registry)
        xml_views = xml_view_generator.generateXML(view_data_list)

        for xml_job in xml_views:
            v_xml.append(xml_job.xml)

        d_xml = XML.SubElement(root, "defaultView")
        d_xml.text = data.get("default-view", views[0]["name"])

        c_xml = XML.SubElement(root, "columns")
        # there is a columns element in a columns element
        c_xml = XML.SubElement(c_xml, "columns")
        columns = data.get("columns", [])

        for column in columns:
            if column in COLUMN_DICT:
                XML.SubElement(c_xml, COLUMN_DICT[column])
            else:
                raise ValueError(
                    "Unsupported column %s is not one of: %s."
                    % (column, ", ".join(COLUMN_DICT.keys()))
                )

        return root



#!/usr/bin/env python
# Copyright (C) 2015 Wayne Warren
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import sys

import jenkins_jobs.cli.subcommand.update as update


logger = logging.getLogger(__name__)


class TestSubCommand(update.UpdateSubCommand):
    def parse_args(self, subparser):
        test = subparser.add_parser("test")

        self.parse_option_recursive_exclude(test)

        self.parse_arg_path(test)
        self.parse_arg_names(test)

        test.add_argument(
            "--config-xml",
            action="store_true",
            dest="config_xml",
            default=False,
            help="use alternative output file layout using config.xml files",
        )
        test.add_argument(
            "-p",
            "--plugin-info",
            dest="plugins_info_path",
            default=None,
            help="path to plugin info YAML file",
        )
        test.add_argument(
            "-o", dest="output_dir", default=sys.stdout, help="path to output XML"
        )

    def execute(self, options, jjb_config):
        if not options.config_xml:
            logger.warning(
                "(Deprecated) The default output behavior of"
                " `jenkins-jobs test` when given the --output"
                " flag will change in JJB 3.0."
                " Instead of writing jobs to OUTPUT/jobname;"
                " they will be written to OUTPUT/jobname/config.xml."
                " The new behavior can be enabled by the passing"
                " `--config-xml` parameter."
            )

        builder, xml_jobs, xml_views = self.make_jobs_and_views_xml(
            jjb_config, options.path, options.names
        )

        builder.update_jobs(
            xml_jobs,
            output=options.output_dir,
            n_workers=1,
            config_xml=options.config_xml,
        )
        builder.update_views(
            xml_views,
            output=options.output_dir,
            n_workers=1,
            config_xml=options.config_xml,
        )
